@@ -41,7 +41,7 @@
 #include "pk/gkr-pk-root-storage.h"
 #endif
 
-#include "pkcs11/gkr-pkcs11-daemon.h"
+#include "pkcs11/gkr-pkcs11-dispatch.h"
 
 #ifdef WITH_SSH
 #include "ssh/gkr-ssh-daemon.h"
@@ -634,8 +634,6 @@ gkr_daemon_complete_initialization(void)
 		return TRUE;
 	}
 	
-	gkr_daemon_dbus_setup ();
-	
 	/* Initialize object storage */
 	if (!gkr_pk_object_storage_initialize ())
 		return FALSE;
@@ -656,10 +654,12 @@ gkr_daemon_complete_initialization(void)
 #endif
 	
 	if (check_run_component ("pkcs11")) {
-		if (!gkr_pkcs11_daemon_setup ())
+		if (!gkr_pkcs11_dispatch_setup ())
 			return FALSE;
 	}
 	
+	gkr_daemon_dbus_setup ();
+	
 	initialization_completed = TRUE;
 	return TRUE;
 }
